from rdkit.Chem import AllChem as Chem # AllChem has most of Chem, and also includes advanced algorithms
import networkx as nx
import matplotlib.pyplot as plt
import sys

class molecule:
	# Initialize input data and variables, run the self.main() function
	def __init__(self, data, filetype, name="molecule", AddHs=True, ThreeD=True, doall=True):
		self.data = data
		self.filetype = filetype
		self.name = name
		self.AddHs = AddHs
		self.ThreeD = ThreeD
		self.mol = None
		self.graph = nx.Graph()
		if doall:
			self.doall()

	# Start with SMILES or MOL format, create RDKit molecule, add Hydrogens, add 3D force field using ETKDG
	def inputMol(self):
		if self.filetype == "smi":
			self.mol = Chem.MolFromSmiles(self.data)
		elif self.filetype == "mol":
			self.mol = Chem.MolFromMolFile(self.data)
		else:
			raise Exception("Accepted data inputs: SMILES string or MOL filepath\nAccepted filetype inputs: 'smi' or 'mol'\nTry again.")
			sys.exit(1)

	# Add hydrogens, make 3D
	def addToMol(self):
		if self.mol == None:
			raise Exception("Molecule was incorrectly created. Cannot add Hydrogens or 3D coordinates.")
			sys.exit(1)
		else:
			if self.AddHs:
				self.mol = Chem.AddHs(self.mol)
			if self.ThreeD:
				Chem.EmbedMolecule(self.mol, randomSeed=0xf00d)

	# Create graph of nodes=atoms and edges=bonds.
	# TODO: Add geometry data, charges
	def graphMol(self):
		if self.mol == None:
			raise Exception("Molecule was incorrectly created. Graph couldn't be created.")
			sys.exit(1)
		else:
			for atom in self.mol.GetAtoms():
				self.graph.add_node(atom.GetIdx(), symbol = atom.GetSymbol())
			for bond in self.mol.GetBonds():
				self.graph.add_edge(bond.GetBeginAtomIdx(), bond.GetEndAtomIdx(), type = bond.GetBondTypeAsDouble())

	# Output in the form of files (.mol file, networkx graph as .graphml) and/or print (.mol data, networkx graph visualized)
	def output(self,file=True, printing=True):
		molBlock = Chem.MolToMolBlock(self.mol)
		if file:
			print(molBlock,file=open(self.name+".mol",'w+'))
			nx.write_graphml(self.graph, self.name+".graphml")
		if printing:
			print(molBlock)
			nx.draw(self.graph, with_labels=True)
			plt.show()

	# input molecule, add Hs/3D coordinates, and create graph
	def doall(self): #does all of the functions, to be run in __init__()
		self.inputMol()
		self.addToMol()
		self.graphMol()



import networkx as nx
from pytket.circuit import Circuit
from pytket.qiskit import tk_to_qiskit
from pytket.routing import Architecture, route
from pytket.passes import SequencePass, DecomposeMultiQubitsIBM, DecomposeSingleQubitsIBM
from pytket.predicates import CompilationUnit

# Load graphml file and convert to NetworkX Graph
def load_graphml(graphml_file):
    return nx.read_graphml(graphml_file)

# Generate the paired CXs representing interactions between atoms
def cx_gen(mol_graph):

    # generate list of tuples containing the edges in the integer-labeled graph
    edges = list(mol_graph.edges) #EDITED TO BE MOL_GRAPH INSTEAD OF MOLE_GRAPH_INTS, SINCE NETWORKX GRAPHS ARE DIFFERENT
    # create a list of INVERTED edges from the original edge list
    inv_edges = [(edge[1], edge[0]) for edge in edges]
    # - Pair together an edge with its inverted counterpart
    # - Store them in one large list
    return [val for pair in zip(edges, inv_edges) for val in pair]

# generate the initial, unrouted circuit with the CXs
def circuit_gen(graph):
    # - generate list of edges along with list of inverted edges
    # - combine the inverted edges with non-inverted ones
    edges = list(graph.edges)
    inv_edges = [(edge[1], edge[0]) for edge in edges]
    cxs = [val for pair in zip(edges, inv_edges) for val in pair]

    # Create circuit and unpack each edge, applying it as a CX
    constraint_test_circuit = Circuit(graph.number_of_nodes())
    for edge in cxs:
        constraint_test_circuit.CX(*edge)

    # return the mapping from atoms to integers and the generated constraint circuit
    return constraint_test_circuit

# convert circuit to IBM Qiskit compatible verison
def circuit_to_ibm(tk_circuit):
    return tk_to_qiskit(tk_circuit)

# print the circuit (implicitly convert to IBM Qiskit supported gates)
def print_circuit(tk_circuit):
    print(tk_to_qiskit(tk_circuit))

# Route the circuit to a given architecture map
def route_circuit(tk_circuit, architecture_map):
    architecture = Architecture(architecture_map)
    routed_circuit = route(tk_circuit, architecture)
    return routed_circuit

# If a subsequent print of the routed circuit fails...
# Invoke the decomposition function with varying "levels" of decomposition
def decompose_circuit(tk_circuit, decomposition_lvl):
    cu = CompilationUnit(tk_circuit)
    # 1 to decompose BRIDGE gates, 2 to decompose as far as possible (includes acceptable SWAPs!)
    if decomposition_lvl == 1:
        seqpass = SequencePass([DecomposeMultiQubitsIBM()])
    elif decomposition_lvl == 2: 
        seqpass = SequencePass([DecomposeMultiQubitsIBM(), DecomposeSingleQubitsIBM()])
    # apply decomposition and return the resulting circuit
    seqpass.apply(cu)
    return cu.circuit



from distutils.core import setup
setup(
  name = 'ALALI',
  packages = ['ALALI'],
  version = '0.0.1',
  license='gpl-3.0',
  description = 'Automated Loading assuming Atomic Level Interactions (ALALI): Mapping molecules to quantum computers',
  author = 'Jack Goon, John Long, Samarth Sandeep',
  author_email = 'jackbgoon@gmail.com',
  url = 'https://github.com/QC-at-Davis/ALALI',
  download_url = 'https://github.com/QC-at-Davis/ALALI/archive/0.0.1.tar.gz', #Need to upload to github before getting download link
  keywords = ['CHEMISTRY', 'QUANTUM', 'QUANTUM CHEMISTRY', 'QUANTUM COMPUTING', 'ATOM', 'QUBIT', 'MAP', 'GRAPH'],   # Keywords that define your package best
  install_requires=[
          'networkx',
          'pytket',
          'pytket-qiskit',
          'matplotlib',
          'rdkit @ https://github.com/rdkit/rdkit/archive/master.zip#egg=rdkit' #not sure on this
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Science/Research',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: GNU General Public License (GPL)',   
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
  ],
)
